# -*- coding: utf-8 -*-

# Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cross Entropy Label Smooth Loss."""
import torch
import torch.nn as nn
from vega.common import ClassFactory, ClassType


@ClassFactory.register(ClassType.LOSS)
class CrossEntropyLabelSmooth(nn.Module):
    """Class of Cross Entropy Label Smooth Loss.

    :param num_classes: number of classes
    :type num_classes: int
    :param epsilon: label smooth coefficient
    :type epsilon: float
    """

    def __init__(self, num_classes, epsilon):
        """Init CrossEntropyLabelSmooth."""
        super(CrossEntropyLabelSmooth, self).__init__()
        self.num_classes = num_classes
        self.epsilon = epsilon
        self.logsoftmax = nn.LogSoftmax(dim=1)

    def forward(self, inputs, targets):
        """Forward function."""
        log_probs = self.logsoftmax(inputs)
        targets = torch.zeros_like(log_probs).scatter_(1, targets.unsqueeze(1), 1)
        targets = (1 - self.epsilon) * targets + self.epsilon / self.num_classes
        loss = (-targets * log_probs).mean(0).sum()
        return loss
